
local MAJOR = "StarLibProperty-1.0" 
local MINOR = 1
assert(LibStub, MAJOR.." requires LibStub") 
local LibProperty = LibStub:NewLibrary(MAJOR, MINOR)
local libEval = "StarLibEvaluator-1.0"
assert(LibStub(libEval), MAJOR .. " requires " .. libEval)
local Evaluator = LibStub(libEval):New()
local errorHandler
assert(LibStub("StarLibError-1.0"), MAJOR .. " requires StarLibError-1.0")

local pool = setmetatable({},{__mode='k'})

if not LibProperty or not Evaluator then return end

if not LibProperty.pool then
	LibProperty.__index = LibProperty
end

-- Create a new StarLibProperty object
-- @name StarLibProperty.New
-- @usage LibProperty:New(env, name, expression, defval)
-- @param visitor A StarLibCore object
-- @param env The execution environment
-- @param name Give your property a name. 
-- @param expression The lua script for this property.
-- @param defval The default value for this property.
-- @return A new StarLibProperty object
function LibProperty:New(visitor, name, expression, defval, errorLevel)
	assert(visitor.environment, "Unable to create new property. Requires an environment")
	assert(name, "Unable to create new property. Requires a name")
	
	if not expression then expression = "" end
	
	local obj = next(pool)

	if obj then
		pool[obj] = nil
	else
		obj = {}
	end

	setmetatable(obj, self)

	obj.visitor = visitor
	
	obj.environment = visitor.environment
	
	obj.expression = expression

	obj.is_valid = true
	
	obj.errorLevel = errorLevel
	
	obj.error = LibStub("StarLibError-1.0"):New(MAJOR, errorLevel)
		
	if expression ~= nil and type(expression) == "string" then
		obj.result = Evaluator.ExecuteCode(env, name, expression, false, defval)
		if obj.result == nil and not expression then
			obj.error:Print(("Property invalid: expression = \"%s\""):format(expression))
			obj.is_valid = false
		end
	elseif expression ~= nil then
		self.error:Print(("Property: <%s> has no expression."):format(name), 1)
		obj.is_valid = false
	end
	
	obj.New = nil	
	
	return obj	
end

-- Delete a property
-- @name StarLibPropoerty.Del
-- @usage LibProperty:Del([prop])
-- @param prop An optional property.
-- @return Nothing
function LibProperty:Del(prop)
	if prop then
		LibProperty.pool[prop] = true
	else
		LibProperty.pool[self] = true
	end
end

-- Evaluate a property's code
-- @name StarLibProperty.Eval
-- @usage object:Eval()
-- @return Nothing
function LibProperty:Eval()
	if not self.is_valid then return -1 end
	
	local update = 1
	
	local old = self.result
	
	self.result = Evaluator.ExecuteCode(self.environment, self.expression)
	
	if old == self.result then
		update = 0
	end
	
	return update
end

-- Return the property's value as a number
-- @name StarLibProperty.P2N
-- @usage object:P2N()
-- @return Nothing
function LibProperty:P2N()
	if not self.is_valid then return 0 end
	if type(self.result) == "number" then
		return self.result
	else
		return tonumber(self.result)
	end
end

-- Return the property's value as a string
-- @name StarLibProperty.P2S
-- @usage object:P2N()
-- @return Nothing
function LibProperty:P2S()
	if not self.is_valid then return "" end
	
	if type(self.result) ~= "number" and type(self.result) ~= "string" then
		return ""
	end
	return ("%s"):format(self.result)
end
